{
 "cells": [
  {
   "cell_type": "markdown",
   "id": "alpine-enemy",
   "metadata": {},
   "source": [
    "# MNIST training A Bigger Model"
   ]
  },
  {
   "cell_type": "code",
   "execution_count": 1,
   "id": "incorrect-radar",
   "metadata": {},
   "outputs": [],
   "source": [
    "import torch\n",
    "import torchvision\n",
    "\n",
    "import numpy as np\n",
    "import math"
   ]
  },
  {
   "cell_type": "code",
   "execution_count": 2,
   "id": "understanding-richards",
   "metadata": {},
   "outputs": [],
   "source": [
    "mnist_trainset = torchvision.datasets.MNIST(root='./data', train=True, download=True, transform=torchvision.transforms.Compose([\n",
    "                               torchvision.transforms.ToTensor(),\n",
    "                               torchvision.transforms.Normalize(\n",
    "                                 (0.1307,), (0.3081,))\n",
    "                             ]))\n",
    "mnist_testset = torchvision.datasets.MNIST(root='./data', train=False, download=True, transform=torchvision.transforms.Compose([\n",
    "                               torchvision.transforms.ToTensor(),\n",
    "                               torchvision.transforms.Normalize(\n",
    "                                 (0.1307,), (0.3081,))\n",
    "                             ]))"
   ]
  },
  {
   "cell_type": "code",
   "execution_count": 3,
   "id": "cognitive-community",
   "metadata": {},
   "outputs": [],
   "source": [
    "## Parameters:\n",
    "n_epochs = 3\n",
    "batch_size_train = 10000\n",
    "batch_size_test = 500\n",
    "log_interval = 500"
   ]
  },
  {
   "cell_type": "code",
   "execution_count": 4,
   "id": "mexican-kelly",
   "metadata": {},
   "outputs": [],
   "source": [
    "train_loader = torch.utils.data.DataLoader(mnist_trainset,batch_size=batch_size_train, shuffle=True)\n",
    "test_loader = torch.utils.data.DataLoader(mnist_testset,batch_size=batch_size_test, shuffle=False)"
   ]
  },
  {
   "cell_type": "code",
   "execution_count": 5,
   "id": "herbal-albania",
   "metadata": {},
   "outputs": [],
   "source": [
    "import papayaclient"
   ]
  },
  {
   "cell_type": "code",
   "execution_count": 6,
   "id": "complex-morning",
   "metadata": {},
   "outputs": [],
   "source": [
    "class TheModel(torch.nn.Module):\n",
    "\n",
    "    def __init__(self):\n",
    "        super(TheModel, self).__init__()\n",
    "\n",
    "        self.conv1 = torch.nn.Conv2d(1, 1, 5)\n",
    "        self.linear1 = torch.nn.Linear((24 * 24), 10)\n",
    "\n",
    "    def forward(self, x):\n",
    "        x1 = self.conv1(x)\n",
    "        x2 = x1.flatten(start_dim = 1)\n",
    "        return self.linear1(x2)"
   ]
  },
  {
   "cell_type": "code",
   "execution_count": 7,
   "id": "announced-convention",
   "metadata": {},
   "outputs": [],
   "source": [
    "clients = []\n",
    "for batchno, (ex_data, ex_labels) in enumerate(train_loader):\n",
    "    clients.append(papayaclient.PapayaClient(dat = ex_data,\n",
    "                                            labs = ex_labels,\n",
    "                                            batch_sz = 500,\n",
    "                                            num_partners = 5,\n",
    "                                            model_class = TheModel,\n",
    "                                            loss_fn = torch.nn.CrossEntropyLoss))\n"
   ]
  },
  {
   "cell_type": "code",
   "execution_count": 8,
   "id": "removed-teens",
   "metadata": {},
   "outputs": [
    {
     "name": "stdout",
     "output_type": "stream",
     "text": [
      "node1888epoch 0 loss 1.5461899042129517\n",
      "node1888epoch 1 loss 0.807489275932312\n",
      "node1888epoch 2 loss 0.6377502679824829\n",
      "node1888epoch 3 loss 0.5672084093093872\n",
      "node1888epoch 4 loss 0.5278582572937012\n",
      "node2786epoch 0 loss 1.9848732948303223\n",
      "node2786epoch 1 loss 1.3048557043075562\n",
      "node2786epoch 2 loss 0.8567324876785278\n",
      "node2786epoch 3 loss 0.6288896203041077\n",
      "node2786epoch 4 loss 0.5104590654373169\n",
      "node3408epoch 0 loss 1.9299290180206299\n",
      "node3408epoch 1 loss 0.9124659895896912\n",
      "node3408epoch 2 loss 0.5893670916557312\n",
      "node3408epoch 3 loss 0.49376410245895386\n",
      "node3408epoch 4 loss 0.4466298222541809\n",
      "node3302epoch 0 loss 1.4667850732803345\n",
      "node3302epoch 1 loss 0.6843005418777466\n",
      "node3302epoch 2 loss 0.5147010684013367\n",
      "node3302epoch 3 loss 0.44649094343185425\n",
      "node3302epoch 4 loss 0.4092641770839691\n",
      "node4282epoch 0 loss 1.3072646856307983\n",
      "node4282epoch 1 loss 0.7020670175552368\n",
      "node4282epoch 2 loss 0.5624675750732422\n",
      "node4282epoch 3 loss 0.5059300065040588\n",
      "node4282epoch 4 loss 0.47413772344589233\n",
      "node1202epoch 0 loss 1.6222590208053589\n",
      "node1202epoch 1 loss 0.6965509653091431\n",
      "node1202epoch 2 loss 0.47517746686935425\n",
      "node1202epoch 3 loss 0.39328184723854065\n",
      "node1202epoch 4 loss 0.35003405809402466\n",
      "node1888epoch 5 loss 0.5022472143173218\n",
      "node1888epoch 6 loss 0.4838199019432068\n",
      "node1888epoch 7 loss 0.46965277194976807\n",
      "node1888epoch 8 loss 0.45827987790107727\n",
      "node1888epoch 9 loss 0.44888386130332947\n",
      "node2786epoch 5 loss 0.44339096546173096\n",
      "node2786epoch 6 loss 0.40240636467933655\n",
      "node2786epoch 7 loss 0.375591516494751\n",
      "node2786epoch 8 loss 0.3569307327270508\n",
      "node2786epoch 9 loss 0.3432456851005554\n",
      "node3408epoch 5 loss 0.417106032371521\n",
      "node3408epoch 6 loss 0.3963075876235962\n",
      "node3408epoch 7 loss 0.38065433502197266\n",
      "node3408epoch 8 loss 0.3683650493621826\n",
      "node3408epoch 9 loss 0.3584255576133728\n",
      "node3302epoch 5 loss 0.3857082724571228\n",
      "node3302epoch 6 loss 0.36939752101898193\n",
      "node3302epoch 7 loss 0.3573715090751648\n",
      "node3302epoch 8 loss 0.34808701276779175\n",
      "node3302epoch 9 loss 0.3406665623188019\n",
      "node4282epoch 5 loss 0.45294925570487976\n",
      "node4282epoch 6 loss 0.4374071955680847\n",
      "node4282epoch 7 loss 0.42530709505081177\n",
      "node4282epoch 8 loss 0.4155017137527466\n",
      "node4282epoch 9 loss 0.40732434391975403\n",
      "node1202epoch 5 loss 0.32317915558815\n",
      "node1202epoch 6 loss 0.30493247509002686\n",
      "node1202epoch 7 loss 0.29178059101104736\n",
      "node1202epoch 8 loss 0.2818772494792938\n",
      "node1202epoch 9 loss 0.2741549611091614\n",
      "node1888epoch 10 loss 0.4409623444080353\n",
      "node1888epoch 11 loss 0.4341803789138794\n",
      "node1888epoch 12 loss 0.4282996952533722\n",
      "node1888epoch 13 loss 0.42314353585243225\n",
      "node1888epoch 14 loss 0.41857704520225525\n",
      "node2786epoch 10 loss 0.3327675461769104\n",
      "node2786epoch 11 loss 0.3244563341140747\n",
      "node2786epoch 12 loss 0.31766989827156067\n",
      "node2786epoch 13 loss 0.31199461221694946\n",
      "node2786epoch 14 loss 0.3071542978286743\n",
      "node3408epoch 10 loss 0.3502052426338196\n",
      "node3408epoch 11 loss 0.34328603744506836\n",
      "node3408epoch 12 loss 0.33737680315971375\n",
      "node3408epoch 13 loss 0.33226656913757324\n",
      "node3408epoch 14 loss 0.32779788970947266\n",
      "node3302epoch 10 loss 0.3345753848552704\n",
      "node3302epoch 11 loss 0.3294681906700134\n",
      "node3302epoch 12 loss 0.32511118054389954\n",
      "node3302epoch 13 loss 0.32133984565734863\n",
      "node3302epoch 14 loss 0.31803518533706665\n",
      "node4282epoch 10 loss 0.4003545939922333\n",
      "node4282epoch 11 loss 0.39431044459342957\n",
      "node4282epoch 12 loss 0.38899290561676025\n",
      "node4282epoch 13 loss 0.38425713777542114\n",
      "node4282epoch 14 loss 0.37999436259269714\n",
      "node1202epoch 10 loss 0.26795637607574463\n",
      "node1202epoch 11 loss 0.26285848021507263\n",
      "node1202epoch 12 loss 0.25857892632484436\n",
      "node1202epoch 13 loss 0.25492382049560547\n",
      "node1202epoch 14 loss 0.2517562806606293\n",
      "node1888epoch 15 loss 0.49509018659591675\n",
      "node1888epoch 16 loss 0.4634162485599518\n",
      "node1888epoch 17 loss 0.44758009910583496\n",
      "node1888epoch 18 loss 0.437000572681427\n",
      "node1888epoch 19 loss 0.42889925837516785\n",
      "node2786epoch 15 loss 0.32510051131248474\n",
      "node2786epoch 16 loss 0.3118307590484619\n",
      "node2786epoch 17 loss 0.30360114574432373\n",
      "node2786epoch 18 loss 0.29781726002693176\n",
      "node2786epoch 19 loss 0.293416291475296\n",
      "node3408epoch 15 loss 0.6490024328231812\n",
      "node3408epoch 16 loss 0.5406443476676941\n",
      "node3408epoch 17 loss 0.49067002534866333\n",
      "node3408epoch 18 loss 0.4579440951347351\n",
      "node3408epoch 19 loss 0.43347156047821045\n",
      "node3302epoch 15 loss 0.44686228036880493\n",
      "node3302epoch 16 loss 0.376604288816452\n",
      "node3302epoch 17 loss 0.3498392105102539\n",
      "node3302epoch 18 loss 0.3351794481277466\n",
      "node3302epoch 19 loss 0.3256184756755829\n",
      "node4282epoch 15 loss 0.47145289182662964\n",
      "node4282epoch 16 loss 0.41765737533569336\n",
      "node4282epoch 17 loss 0.3951617479324341\n",
      "node4282epoch 18 loss 0.382162868976593\n",
      "node4282epoch 19 loss 0.37335315346717834\n",
      "node1202epoch 15 loss 0.5093368887901306\n",
      "node1202epoch 16 loss 0.3888815641403198\n",
      "node1202epoch 17 loss 0.3389842212200165\n",
      "node1202epoch 18 loss 0.3107586205005646\n",
      "node1202epoch 19 loss 0.2925325334072113\n",
      "node1888epoch 20 loss 0.45341265201568604\n",
      "node1888epoch 21 loss 0.4388343393802643\n",
      "node1888epoch 22 loss 0.4293522536754608\n",
      "node1888epoch 23 loss 0.422124981880188\n",
      "node1888epoch 24 loss 0.41618505120277405\n",
      "node2786epoch 20 loss 0.33227667212486267\n",
      "node2786epoch 21 loss 0.3064635992050171\n",
      "node2786epoch 22 loss 0.2940453886985779\n",
      "node2786epoch 23 loss 0.2867438495159149\n",
      "node2786epoch 24 loss 0.28193575143814087\n",
      "node3408epoch 20 loss 0.47568845748901367\n",
      "node3408epoch 21 loss 0.42927348613739014\n",
      "node3408epoch 22 loss 0.4002789855003357\n",
      "node3408epoch 23 loss 0.3798483908176422\n",
      "node3408epoch 24 loss 0.36469921469688416\n",
      "node3302epoch 20 loss 0.34759217500686646\n",
      "node3302epoch 21 loss 0.33178290724754333\n",
      "node3302epoch 22 loss 0.3221155107021332\n",
      "node3302epoch 23 loss 0.3153358995914459\n",
      "node3302epoch 24 loss 0.31022170186042786\n",
      "node4282epoch 20 loss 0.43591442704200745\n",
      "node4282epoch 21 loss 0.4000609517097473\n",
      "node4282epoch 22 loss 0.38178786635398865\n",
      "node4282epoch 23 loss 0.3702560365200043\n",
      "node4282epoch 24 loss 0.3621349632740021\n",
      "node1202epoch 20 loss 0.5307437777519226\n",
      "node1202epoch 21 loss 0.4502067565917969\n",
      "node1202epoch 22 loss 0.40706634521484375\n",
      "node1202epoch 23 loss 0.37812310457229614\n",
      "node1202epoch 24 loss 0.3566901981830597\n",
      "node1888epoch 25 loss 0.4463861584663391\n",
      "node1888epoch 26 loss 0.4338056445121765\n",
      "node1888epoch 27 loss 0.42522209882736206\n",
      "node1888epoch 28 loss 0.4185144305229187\n",
      "node1888epoch 29 loss 0.41293591260910034\n",
      "node2786epoch 25 loss 0.28731581568717957\n",
      "node2786epoch 26 loss 0.28184637427330017\n",
      "node2786epoch 27 loss 0.2781100869178772\n",
      "node2786epoch 28 loss 0.2753494679927826\n",
      "node2786epoch 29 loss 0.2732083201408386\n",
      "node3408epoch 25 loss 0.3539830148220062\n",
      "node3408epoch 26 loss 0.3434596061706543\n",
      "node3408epoch 27 loss 0.33556216955184937\n",
      "node3408epoch 28 loss 0.32936471700668335\n",
      "node3408epoch 29 loss 0.3243376910686493\n",
      "node3302epoch 25 loss 0.32053524255752563\n",
      "node3302epoch 26 loss 0.31327512860298157\n",
      "node3302epoch 27 loss 0.30801671743392944\n",
      "node3302epoch 28 loss 0.3039668798446655\n",
      "node3302epoch 29 loss 0.30071452260017395\n",
      "node4282epoch 25 loss 0.37698060274124146\n",
      "node4282epoch 26 loss 0.3654605746269226\n",
      "node4282epoch 27 loss 0.3575025796890259\n",
      "node4282epoch 28 loss 0.3515063226222992\n",
      "node4282epoch 29 loss 0.3467514216899872\n",
      "node1202epoch 25 loss 0.394527405500412\n",
      "node1202epoch 26 loss 0.35917940735816956\n",
      "node1202epoch 27 loss 0.33541062474250793\n",
      "node1202epoch 28 loss 0.31790921092033386\n",
      "node1202epoch 29 loss 0.30441731214523315\n",
      "node1888epoch 30 loss 0.4115205407142639\n",
      "node1888epoch 31 loss 0.4061899781227112\n",
      "node1888epoch 32 loss 0.40220654010772705\n",
      "node1888epoch 33 loss 0.3987683951854706\n",
      "node1888epoch 34 loss 0.3956773281097412\n",
      "node2786epoch 30 loss 0.27604377269744873\n",
      "node2786epoch 31 loss 0.27343058586120605\n",
      "node2786epoch 32 loss 0.27147144079208374\n",
      "node2786epoch 33 loss 0.26989781856536865\n",
      "node2786epoch 34 loss 0.2685846984386444\n",
      "node3408epoch 30 loss 0.33101433515548706\n",
      "node3408epoch 31 loss 0.32451191544532776\n",
      "node3408epoch 32 loss 0.3195887506008148\n",
      "node3408epoch 33 loss 0.3156214654445648\n",
      "node3408epoch 34 loss 0.3123023808002472\n",
      "node3302epoch 30 loss 0.3114315867424011\n",
      "node3302epoch 31 loss 0.30572450160980225\n",
      "node3302epoch 32 loss 0.3015368580818176\n",
      "node3302epoch 33 loss 0.2982677221298218\n",
      "node3302epoch 34 loss 0.2956050634384155\n",
      "node4282epoch 30 loss 0.3471895158290863\n",
      "node4282epoch 31 loss 0.342898428440094\n",
      "node4282epoch 32 loss 0.3394930958747864\n",
      "node4282epoch 33 loss 0.3365868330001831\n",
      "node4282epoch 34 loss 0.3340296447277069\n",
      "node1202epoch 30 loss 0.28304529190063477\n",
      "node1202epoch 31 loss 0.273284375667572\n",
      "node1202epoch 32 loss 0.2662597596645355\n",
      "node1202epoch 33 loss 0.2608601748943329\n",
      "node1202epoch 34 loss 0.25653794407844543\n",
      "node1888epoch 35 loss 0.40062206983566284\n",
      "node1888epoch 36 loss 0.3968391418457031\n",
      "node1888epoch 37 loss 0.3938008248806\n",
      "node1888epoch 38 loss 0.3910719156265259\n",
      "node1888epoch 39 loss 0.3885616064071655\n",
      "node2786epoch 35 loss 0.2716084122657776\n",
      "node2786epoch 36 loss 0.2694401741027832\n",
      "node2786epoch 37 loss 0.26783478260040283\n",
      "node2786epoch 38 loss 0.2665363848209381\n",
      "node2786epoch 39 loss 0.2654415965080261\n",
      "node3408epoch 35 loss 0.31628111004829407\n",
      "node3408epoch 36 loss 0.31186771392822266\n",
      "node3408epoch 37 loss 0.3085353672504425\n",
      "node3408epoch 38 loss 0.3058021664619446\n",
      "node3408epoch 39 loss 0.3034597337245941\n",
      "node3302epoch 35 loss 0.30227893590927124\n",
      "node3302epoch 36 loss 0.2980383634567261\n",
      "node3302epoch 37 loss 0.295043021440506\n",
      "node3302epoch 38 loss 0.2927111089229584\n",
      "node3302epoch 39 loss 0.2907745838165283\n",
      "node4282epoch 35 loss 0.3329603970050812\n",
      "node4282epoch 36 loss 0.3305725157260895\n",
      "node4282epoch 37 loss 0.3286473751068115\n",
      "node4282epoch 38 loss 0.32690247893333435\n",
      "node4282epoch 39 loss 0.32526934146881104\n",
      "node1202epoch 35 loss 0.2539554834365845\n",
      "node1202epoch 36 loss 0.25020915269851685\n",
      "node1202epoch 37 loss 0.2473328560590744\n",
      "node1202epoch 38 loss 0.24497662484645844\n",
      "node1202epoch 39 loss 0.2429785132408142\n",
      "node1888epoch 40 loss 0.3929089307785034\n",
      "node1888epoch 41 loss 0.3893277049064636\n",
      "node1888epoch 42 loss 0.38663309812545776\n",
      "node1888epoch 43 loss 0.38427600264549255\n",
      "node1888epoch 44 loss 0.38212475180625916\n",
      "node2786epoch 40 loss 0.2688120901584625\n",
      "node2786epoch 41 loss 0.2668490409851074\n",
      "node2786epoch 42 loss 0.2654011845588684\n",
      "node2786epoch 43 loss 0.26422667503356934\n",
      "node2786epoch 44 loss 0.2632324993610382\n",
      "node3408epoch 40 loss 0.307399183511734\n",
      "node3408epoch 41 loss 0.3038480579853058\n",
      "node3408epoch 42 loss 0.3012009263038635\n",
      "node3408epoch 43 loss 0.2990320920944214\n",
      "node3408epoch 44 loss 0.29716479778289795\n",
      "node3302epoch 40 loss 0.2994071841239929\n",
      "node3302epoch 41 loss 0.29533401131629944\n",
      "node3302epoch 42 loss 0.29258015751838684\n",
      "node3302epoch 43 loss 0.29046425223350525\n",
      "node3302epoch 44 loss 0.2887042164802551\n",
      "node4282epoch 40 loss 0.32434290647506714\n",
      "node4282epoch 41 loss 0.3228031098842621\n",
      "node4282epoch 42 loss 0.32148975133895874\n",
      "node4282epoch 43 loss 0.3202255368232727\n",
      "node4282epoch 44 loss 0.31898394227027893\n",
      "node1202epoch 40 loss 0.24552510678768158\n",
      "node1202epoch 41 loss 0.24277664721012115\n",
      "node1202epoch 42 loss 0.24071986973285675\n",
      "node1202epoch 43 loss 0.23904964327812195\n",
      "node1202epoch 44 loss 0.23763015866279602\n",
      "node1888epoch 45 loss 0.3860745131969452\n",
      "node1888epoch 46 loss 0.38304537534713745\n",
      "node1888epoch 47 loss 0.3807604908943176\n",
      "node1888epoch 48 loss 0.3787422776222229\n",
      "node1888epoch 49 loss 0.3768891394138336\n",
      "node2786epoch 45 loss 0.2666627764701843\n",
      "node2786epoch 46 loss 0.2646707594394684\n",
      "node2786epoch 47 loss 0.26324281096458435\n",
      "node2786epoch 48 loss 0.26210924983024597\n",
      "node2786epoch 49 loss 0.2611669898033142\n",
      "node3408epoch 45 loss 0.3019930422306061\n",
      "node3408epoch 46 loss 0.2986372113227844\n",
      "node3408epoch 47 loss 0.2962087392807007\n",
      "node3408epoch 48 loss 0.2942563593387604\n",
      "node3408epoch 49 loss 0.29259538650512695\n",
      "node3302epoch 45 loss 0.2964019179344177\n",
      "node3302epoch 46 loss 0.2928466200828552\n",
      "node3302epoch 47 loss 0.2904180586338043\n",
      "node3302epoch 48 loss 0.28853553533554077\n",
      "node3302epoch 49 loss 0.28695210814476013\n",
      "node4282epoch 45 loss 0.32067227363586426\n",
      "node4282epoch 46 loss 0.3188387453556061\n",
      "node4282epoch 47 loss 0.31746697425842285\n",
      "node4282epoch 48 loss 0.31622567772865295\n",
      "node4282epoch 49 loss 0.3150428831577301\n",
      "node1202epoch 45 loss 0.23997628688812256\n",
      "node1202epoch 46 loss 0.2378583699464798\n",
      "node1202epoch 47 loss 0.236273393034935\n",
      "node1202epoch 48 loss 0.23497889935970306\n",
      "node1202epoch 49 loss 0.2338719516992569\n",
      "node1888epoch 50 loss 0.3799128532409668\n",
      "node1888epoch 51 loss 0.37706589698791504\n",
      "node1888epoch 52 loss 0.37509140372276306\n",
      "node1888epoch 53 loss 0.3733851909637451\n",
      "node1888epoch 54 loss 0.3718212842941284\n",
      "node2786epoch 50 loss 0.2654806077480316\n",
      "node2786epoch 51 loss 0.2636384069919586\n",
      "node2786epoch 52 loss 0.2622569501399994\n",
      "node2786epoch 53 loss 0.2611297369003296\n",
      "node2786epoch 54 loss 0.26018026471138\n",
      "node3408epoch 50 loss 0.2990410625934601\n",
      "node3408epoch 51 loss 0.2956397235393524\n",
      "node3408epoch 52 loss 0.2931853234767914\n",
      "node3408epoch 53 loss 0.29123952984809875\n",
      "node3408epoch 54 loss 0.28960689902305603\n",
      "node3302epoch 50 loss 0.29446151852607727\n",
      "node3302epoch 51 loss 0.29084327816963196\n",
      "node3302epoch 52 loss 0.2884739935398102\n",
      "node3302epoch 53 loss 0.286662757396698\n",
      "node3302epoch 54 loss 0.2851453721523285\n",
      "node4282epoch 50 loss 0.3152286410331726\n",
      "node4282epoch 51 loss 0.31408581137657166\n",
      "node4282epoch 52 loss 0.31309035420417786\n",
      "node4282epoch 53 loss 0.3120960593223572\n",
      "node4282epoch 54 loss 0.31109291315078735\n",
      "node1202epoch 50 loss 0.2377670407295227\n",
      "node1202epoch 51 loss 0.23553507030010223\n",
      "node1202epoch 52 loss 0.2339804768562317\n",
      "node1202epoch 53 loss 0.23276576399803162\n",
      "node1202epoch 54 loss 0.23175452649593353\n",
      "node1888epoch 55 loss 0.3775845468044281\n",
      "node1888epoch 56 loss 0.3746410310268402\n",
      "node1888epoch 57 loss 0.3727017641067505\n",
      "node1888epoch 58 loss 0.37105098366737366\n",
      "node1888epoch 59 loss 0.3695434629917145\n",
      "node2786epoch 55 loss 0.26328378915786743\n",
      "node2786epoch 56 loss 0.26152506470680237\n",
      "node2786epoch 57 loss 0.26025059819221497\n",
      "node2786epoch 58 loss 0.2592243254184723\n",
      "node2786epoch 59 loss 0.25836408138275146\n",
      "node3408epoch 55 loss 0.29503124952316284\n",
      "node3408epoch 56 loss 0.29182296991348267\n",
      "node3408epoch 57 loss 0.2895401120185852\n",
      "node3408epoch 58 loss 0.2877533435821533\n",
      "node3408epoch 59 loss 0.28627005219459534\n",
      "node3302epoch 55 loss 0.2934997081756592\n",
      "node3302epoch 56 loss 0.2897948920726776\n",
      "node3302epoch 57 loss 0.2874014675617218\n",
      "node3302epoch 58 loss 0.2855971157550812\n",
      "node3302epoch 59 loss 0.2840995788574219\n",
      "node4282epoch 55 loss 0.31343695521354675\n",
      "node4282epoch 56 loss 0.3117128610610962\n",
      "node4282epoch 57 loss 0.31046992540359497\n",
      "node4282epoch 58 loss 0.30936354398727417\n",
      "node4282epoch 59 loss 0.3083169460296631\n",
      "node1202epoch 55 loss 0.2346033751964569\n",
      "node1202epoch 56 loss 0.23276641964912415\n",
      "node1202epoch 57 loss 0.23144122958183289\n",
      "node1202epoch 58 loss 0.23038537800312042\n",
      "node1202epoch 59 loss 0.22949910163879395\n",
      "node1888epoch 60 loss 0.37378981709480286\n",
      "node1888epoch 61 loss 0.3711438775062561\n",
      "node1888epoch 62 loss 0.3692989647388458\n",
      "node1888epoch 63 loss 0.3677237927913666\n",
      "node1888epoch 64 loss 0.36630091071128845\n",
      "node2786epoch 60 loss 0.26327335834503174\n",
      "node2786epoch 61 loss 0.26146525144577026\n",
      "node2786epoch 62 loss 0.2600889801979065\n",
      "node2786epoch 63 loss 0.25896963477134705\n",
      "node2786epoch 64 loss 0.25803637504577637\n",
      "node3408epoch 60 loss 0.2922360301017761\n",
      "node3408epoch 61 loss 0.28899475932121277\n",
      "node3408epoch 62 loss 0.2867927551269531\n",
      "node3408epoch 63 loss 0.28509727120399475\n",
      "node3408epoch 64 loss 0.28369811177253723\n",
      "node3302epoch 60 loss 0.291280061006546\n",
      "node3302epoch 61 loss 0.28807470202445984\n",
      "node3302epoch 62 loss 0.285942405462265\n",
      "node3302epoch 63 loss 0.28429853916168213\n",
      "node3302epoch 64 loss 0.28291112184524536\n",
      "node4282epoch 60 loss 0.31053170561790466\n",
      "node4282epoch 61 loss 0.3089609444141388\n",
      "node4282epoch 62 loss 0.3077877163887024\n",
      "node4282epoch 63 loss 0.30672886967658997\n",
      "node4282epoch 64 loss 0.30572283267974854\n",
      "node1202epoch 60 loss 0.23283590376377106\n",
      "node1202epoch 61 loss 0.230974480509758\n",
      "node1202epoch 62 loss 0.22971874475479126\n",
      "node1202epoch 63 loss 0.22875696420669556\n",
      "node1202epoch 64 loss 0.227966770529747\n",
      "node1888epoch 65 loss 0.3718826174736023\n",
      "node1888epoch 66 loss 0.36893877387046814\n",
      "node1888epoch 67 loss 0.36708390712738037\n",
      "node1888epoch 68 loss 0.365543931722641\n",
      "node1888epoch 69 loss 0.36415907740592957\n",
      "node2786epoch 65 loss 0.26155388355255127\n",
      "node2786epoch 66 loss 0.25976094603538513\n",
      "node2786epoch 67 loss 0.25842398405075073\n",
      "node2786epoch 68 loss 0.25735852122306824\n",
      "node2786epoch 69 loss 0.25648602843284607\n",
      "node3408epoch 65 loss 0.2889217138290405\n",
      "node3408epoch 66 loss 0.28589928150177\n",
      "node3408epoch 67 loss 0.283867746591568\n",
      "node3408epoch 68 loss 0.2823230028152466\n",
      "node3408epoch 69 loss 0.28106069564819336\n",
      "node3302epoch 65 loss 0.2912961542606354\n",
      "node3302epoch 66 loss 0.28771698474884033\n",
      "node3302epoch 67 loss 0.28543463349342346\n",
      "node3302epoch 68 loss 0.28370901942253113\n",
      "node3302epoch 69 loss 0.2822689414024353\n",
      "node4282epoch 65 loss 0.307779461145401\n",
      "node4282epoch 66 loss 0.30605369806289673\n",
      "node4282epoch 67 loss 0.30484336614608765\n",
      "node4282epoch 68 loss 0.30379390716552734\n",
      "node4282epoch 69 loss 0.30281928181648254\n",
      "node1202epoch 65 loss 0.2312219887971878\n",
      "node1202epoch 66 loss 0.22945955395698547\n",
      "node1202epoch 67 loss 0.22825635969638824\n",
      "node1202epoch 68 loss 0.22733430564403534\n",
      "node1202epoch 69 loss 0.22658087313175201\n",
      "node1888epoch 70 loss 0.3698602616786957\n",
      "node1888epoch 71 loss 0.36695578694343567\n",
      "node1888epoch 72 loss 0.36512550711631775\n",
      "node1888epoch 73 loss 0.36360785365104675\n",
      "node1888epoch 74 loss 0.36224693059921265\n",
      "node2786epoch 70 loss 0.26041802763938904\n",
      "node2786epoch 71 loss 0.25863853096961975\n",
      "node2786epoch 72 loss 0.2573189437389374\n",
      "node2786epoch 73 loss 0.25627660751342773\n",
      "node2786epoch 74 loss 0.25543123483657837\n",
      "node3408epoch 70 loss 0.2887267768383026\n",
      "node3408epoch 71 loss 0.28526633977890015\n",
      "node3408epoch 72 loss 0.2829481363296509\n",
      "node3408epoch 73 loss 0.2812167704105377\n",
      "node3408epoch 74 loss 0.27983108162879944\n",
      "node3302epoch 70 loss 0.28957492113113403\n",
      "node3302epoch 71 loss 0.2860984206199646\n",
      "node3302epoch 72 loss 0.28393855690956116\n",
      "node3302epoch 73 loss 0.2823293209075928\n",
      "node3302epoch 74 loss 0.28099438548088074\n",
      "node4282epoch 70 loss 0.3058343827724457\n",
      "node4282epoch 71 loss 0.3040216267108917\n",
      "node4282epoch 72 loss 0.3027768135070801\n",
      "node4282epoch 73 loss 0.3017142713069916\n",
      "node4282epoch 74 loss 0.3007391393184662\n",
      "node1202epoch 70 loss 0.2304108738899231\n",
      "node1202epoch 71 loss 0.22859713435173035\n",
      "node1202epoch 72 loss 0.2273990958929062\n",
      "node1202epoch 73 loss 0.22650133073329926\n",
      "node1202epoch 74 loss 0.2257779836654663\n",
      "node1888epoch 75 loss 0.3667471408843994\n",
      "node1888epoch 76 loss 0.36415401101112366\n",
      "node1888epoch 77 loss 0.36242973804473877\n",
      "node1888epoch 78 loss 0.36099937558174133\n",
      "node1888epoch 79 loss 0.35973095893859863\n",
      "node2786epoch 75 loss 0.26021674275398254\n",
      "node2786epoch 76 loss 0.25841599702835083\n",
      "node2786epoch 77 loss 0.2570604383945465\n",
      "node2786epoch 78 loss 0.25598540902137756\n",
      "node2786epoch 79 loss 0.2551126778125763\n",
      "node3408epoch 75 loss 0.28706037998199463\n",
      "node3408epoch 76 loss 0.2835080027580261\n",
      "node3408epoch 77 loss 0.281212717294693\n",
      "node3408epoch 78 loss 0.27952301502227783\n",
      "node3408epoch 79 loss 0.2781790494918823\n",
      "node3302epoch 75 loss 0.28854334354400635\n",
      "node3302epoch 76 loss 0.2853190004825592\n",
      "node3302epoch 77 loss 0.28324422240257263\n",
      "node3302epoch 78 loss 0.28166869282722473\n",
      "node3302epoch 79 loss 0.2803479731082916\n",
      "node4282epoch 75 loss 0.30315548181533813\n",
      "node4282epoch 76 loss 0.301567941904068\n",
      "node4282epoch 77 loss 0.30042964220046997\n",
      "node4282epoch 78 loss 0.29943352937698364\n",
      "node4282epoch 79 loss 0.2985078692436218\n",
      "node1202epoch 75 loss 0.2297283560037613\n",
      "node1202epoch 76 loss 0.22785067558288574\n",
      "node1202epoch 77 loss 0.2266109436750412\n",
      "node1202epoch 78 loss 0.22568783164024353\n",
      "node1202epoch 79 loss 0.22495107352733612\n",
      "node1888epoch 80 loss 0.3648010492324829\n",
      "node1888epoch 81 loss 0.36229628324508667\n",
      "node1888epoch 82 loss 0.3606383204460144\n",
      "node1888epoch 83 loss 0.3592694103717804\n",
      "node1888epoch 84 loss 0.3580597937107086\n",
      "node2786epoch 80 loss 0.25890058279037476\n",
      "node2786epoch 81 loss 0.25705453753471375\n",
      "node2786epoch 82 loss 0.2557581961154938\n",
      "node2786epoch 83 loss 0.2547573447227478\n",
      "node2786epoch 84 loss 0.25395333766937256\n",
      "node3408epoch 80 loss 0.2850150167942047\n",
      "node3408epoch 81 loss 0.2816333472728729\n",
      "node3408epoch 82 loss 0.27946630120277405\n",
      "node3408epoch 83 loss 0.27787479758262634\n",
      "node3408epoch 84 loss 0.27660876512527466\n",
      "node3302epoch 80 loss 0.28908485174179077\n",
      "node3302epoch 81 loss 0.28551551699638367\n",
      "node3302epoch 82 loss 0.2832712233066559\n",
      "node3302epoch 83 loss 0.281579852104187\n",
      "node3302epoch 84 loss 0.2801724076271057\n",
      "node4282epoch 80 loss 0.3009029030799866\n",
      "node4282epoch 81 loss 0.29928556084632874\n",
      "node4282epoch 82 loss 0.2981615960597992\n",
      "node4282epoch 83 loss 0.2971959412097931\n",
      "node4282epoch 84 loss 0.29630640149116516\n",
      "node1202epoch 80 loss 0.22859089076519012\n",
      "node1202epoch 81 loss 0.226896271109581\n",
      "node1202epoch 82 loss 0.22576017677783966\n",
      "node1202epoch 83 loss 0.22490815818309784\n",
      "node1202epoch 84 loss 0.22422590851783752\n",
      "node1888epoch 85 loss 0.3643108308315277\n",
      "node1888epoch 86 loss 0.3615228235721588\n",
      "node1888epoch 87 loss 0.359812468290329\n",
      "node1888epoch 88 loss 0.3584159314632416\n",
      "node1888epoch 89 loss 0.35717684030532837\n",
      "node2786epoch 85 loss 0.25901710987091064\n",
      "node2786epoch 86 loss 0.2572079598903656\n",
      "node2786epoch 87 loss 0.2558390498161316\n",
      "node2786epoch 88 loss 0.2547614574432373\n",
      "node2786epoch 89 loss 0.2538954019546509\n",
      "node3408epoch 85 loss 0.283206969499588\n",
      "node3408epoch 86 loss 0.279865562915802\n",
      "node3408epoch 87 loss 0.2777383327484131\n",
      "node3408epoch 88 loss 0.27619749307632446\n",
      "node3408epoch 89 loss 0.2749873697757721\n",
      "node3302epoch 85 loss 0.2886727452278137\n",
      "node3302epoch 86 loss 0.2849300801753998\n",
      "node3302epoch 87 loss 0.28264203667640686\n",
      "node3302epoch 88 loss 0.2809436321258545\n",
      "node3302epoch 89 loss 0.27954351902008057\n",
      "node4282epoch 85 loss 0.29818835854530334\n",
      "node4282epoch 86 loss 0.2969956398010254\n",
      "node4282epoch 87 loss 0.2960343062877655\n",
      "node4282epoch 88 loss 0.29514315724372864\n",
      "node4282epoch 89 loss 0.2942929267883301\n",
      "node1202epoch 85 loss 0.22810031473636627\n",
      "node1202epoch 86 loss 0.2263496071100235\n",
      "node1202epoch 87 loss 0.22518585622310638\n",
      "node1202epoch 88 loss 0.2243177890777588\n",
      "node1202epoch 89 loss 0.22362712025642395\n",
      "node1888epoch 90 loss 0.3631635308265686\n",
      "node1888epoch 91 loss 0.360343337059021\n",
      "node1888epoch 92 loss 0.3586117625236511\n",
      "node1888epoch 93 loss 0.35720276832580566\n",
      "node1888epoch 94 loss 0.35595858097076416\n",
      "node2786epoch 90 loss 0.2572856545448303\n",
      "node2786epoch 91 loss 0.255623996257782\n",
      "node2786epoch 92 loss 0.2544323801994324\n",
      "node2786epoch 93 loss 0.2534979581832886\n",
      "node2786epoch 94 loss 0.2527404725551605\n",
      "node3408epoch 90 loss 0.28147342801094055\n",
      "node3408epoch 91 loss 0.27832770347595215\n",
      "node3408epoch 92 loss 0.27628231048583984\n",
      "node3408epoch 93 loss 0.27479663491249084\n",
      "node3408epoch 94 loss 0.27363330125808716\n",
      "node3302epoch 90 loss 0.28856897354125977\n",
      "node3302epoch 91 loss 0.2848595976829529\n",
      "node3302epoch 92 loss 0.2825697064399719\n",
      "node3302epoch 93 loss 0.2808803617954254\n",
      "node3302epoch 94 loss 0.279491662979126\n",
      "node4282epoch 90 loss 0.2964078485965729\n",
      "node4282epoch 91 loss 0.2950185537338257\n",
      "node4282epoch 92 loss 0.29401424527168274\n",
      "node4282epoch 93 loss 0.2931325435638428\n",
      "node4282epoch 94 loss 0.292310893535614\n",
      "node1202epoch 90 loss 0.2268470674753189\n",
      "node1202epoch 91 loss 0.22527539730072021\n",
      "node1202epoch 92 loss 0.2242085039615631\n",
      "node1202epoch 93 loss 0.2234015017747879\n",
      "node1202epoch 94 loss 0.22275537252426147\n",
      "node1888epoch 95 loss 0.3608856499195099\n",
      "node1888epoch 96 loss 0.35831716656684875\n",
      "node1888epoch 97 loss 0.3566399812698364\n",
      "node1888epoch 98 loss 0.3552776277065277\n",
      "node1888epoch 99 loss 0.35409218072891235\n",
      "node2786epoch 95 loss 0.25813233852386475\n",
      "node2786epoch 96 loss 0.25624752044677734\n",
      "node2786epoch 97 loss 0.2548363506793976\n",
      "node2786epoch 98 loss 0.25373220443725586\n",
      "node2786epoch 99 loss 0.2528506815433502\n",
      "node3408epoch 95 loss 0.2796116769313812\n",
      "node3408epoch 96 loss 0.2765235900878906\n",
      "node3408epoch 97 loss 0.2745947241783142\n",
      "node3408epoch 98 loss 0.2732130289077759\n",
      "node3408epoch 99 loss 0.2721341550350189\n",
      "node3302epoch 95 loss 0.289368599653244\n",
      "node3302epoch 96 loss 0.28559592366218567\n",
      "node3302epoch 97 loss 0.28321489691734314\n",
      "node3302epoch 98 loss 0.2814217209815979\n",
      "node3302epoch 99 loss 0.27993401885032654\n",
      "node4282epoch 95 loss 0.29470372200012207\n",
      "node4282epoch 96 loss 0.29350903630256653\n",
      "node4282epoch 97 loss 0.29255375266075134\n",
      "node4282epoch 98 loss 0.2916756868362427\n",
      "node4282epoch 99 loss 0.29084330797195435\n",
      "node1202epoch 95 loss 0.22645430266857147\n",
      "node1202epoch 96 loss 0.2247922420501709\n",
      "node1202epoch 97 loss 0.22368121147155762\n",
      "node1202epoch 98 loss 0.222855344414711\n",
      "node1202epoch 99 loss 0.22220300137996674\n"
     ]
    }
   ],
   "source": [
    "## Train the Nodes\n",
    "num_epochs_total = 100\n",
    "num_epochs_per_swap = 5\n",
    "num_times = (num_epochs_total // num_epochs_per_swap)\n",
    "for i in range(0, num_times):\n",
    "    for n in clients:\n",
    "        for j in range(0, num_epochs_per_swap):\n",
    "            n.model_train_epoch()\n",
    "            # print(n.logs['stringy'][n.epochs_trained - 1])\n",
    "    if i > 1 and i < num_times - 1 :\n",
    "        for n in clients:\n",
    "            n.select_partners(3)\n",
    "        for n in clients:\n",
    "            for i in range(0, 4) :\n",
    "                n.update_partner_weights()\n",
    "            n.average_partners()"
   ]
  },
  {
   "cell_type": "code",
   "execution_count": 9,
   "id": "accredited-cassette",
   "metadata": {},
   "outputs": [
    {
     "name": "stdout",
     "output_type": "stream",
     "text": [
      "node1888epoch 99 loss 0.35409218072891235\n",
      "node2786epoch 99 loss 0.2528506815433502\n",
      "node3408epoch 99 loss 0.2721341550350189\n",
      "node3302epoch 99 loss 0.27993401885032654\n",
      "node4282epoch 99 loss 0.29084330797195435\n",
      "node1202epoch 99 loss 0.22220300137996674\n"
     ]
    }
   ],
   "source": [
    "for c in clients :\n",
    "    print(c.logs['stringy'][99])"
   ]
  },
  {
   "cell_type": "code",
   "execution_count": 10,
   "id": "certain-gamma",
   "metadata": {},
   "outputs": [],
   "source": [
    "accuracies = {}\n",
    "with torch.no_grad():\n",
    "    for i in clients :\n",
    "        accuracies_node = []\n",
    "        for batchno, (ex_data, ex_labels) in enumerate(test_loader) :\n",
    "            accuracies_node.append(((i.model.forward(ex_data).argmax(dim = 1) == ex_labels).float().mean()).item())\n",
    "        accuracies[i.node_id] = np.array(accuracies_node).mean()"
   ]
  },
  {
   "cell_type": "code",
   "execution_count": 11,
   "id": "periodic-brief",
   "metadata": {},
   "outputs": [
    {
     "data": {
      "text/plain": [
       "{1888: 0.9180999994277954,\n",
       " 2786: 0.9181000024080277,\n",
       " 3408: 0.9206999987363815,\n",
       " 3302: 0.9205000013113022,\n",
       " 4282: 0.9197999954223632,\n",
       " 1202: 0.918299999833107}"
      ]
     },
     "execution_count": 11,
     "metadata": {},
     "output_type": "execute_result"
    }
   ],
   "source": [
    "accuracies"
   ]
  },
  {
   "cell_type": "code",
   "execution_count": null,
   "id": "spatial-biotechnology",
   "metadata": {},
   "outputs": [],
   "source": []
  }
 ],
 "metadata": {
  "kernelspec": {
   "display_name": "Python 3",
   "language": "python",
   "name": "python3"
  },
  "language_info": {
   "codemirror_mode": {
    "name": "ipython",
    "version": 3
   },
   "file_extension": ".py",
   "mimetype": "text/x-python",
   "name": "python",
   "nbconvert_exporter": "python",
   "pygments_lexer": "ipython3",
   "version": "3.8.3"
  }
 },
 "nbformat": 4,
 "nbformat_minor": 5
}
